# ==============
# File: morse.py
# By: Jack Holdsworth - 20023933
# Use: A copy of morse from worksheet 2 part 1
# Written on: Python 3.8
# ==============

class node:
    def __init__(self, value, left=None, right=None):
        self.value = value # Contains itself value
        self.leftChildNode = left # a node object for the left default is None
        self.rightChildNode = right # a node object to the right default 
        

    def setLeftChildNode(self,nodes):
        # setter for left child
        self.leftChildNode = nodes

    def setRightChildNode(self,nodes):
        # setter for right child
        self.rightChildNode = nodes

    def hasLeftChildNode(self):
        # getter for left child
        if self.leftChildNode != None:
            return True
        else:
            return False

    def hasRightChildNode(self):
        # getter for right child
        if self.rightChildNode != None:
            return True
        else:
            return False
    
    def printNode(self, level=1,direction="S"):
        # prints child nodes with directions
        
        # sets spaces for node level. E.I. a child with 3 parents will have 3 spaces
        for x in range(level):
            print(" ", end = '')
        print(direction + " - " + str(self.value)) # Either (L)eft child or (R)ight child
        
        # If has child then call recursion
        if self.hasLeftChildNode() == True:  
            self.leftChildNode.printNode((level+1),"L")
        if self.hasRightChildNode() == True:
            self.rightChildNode.printNode((level+1), "R")

    def words(self,values):
        # a method to splits morse words into an array of morse words and decode them.
        
        if values: # stops recursion when NoneType is returned.
            x = values.split(" ") # split string into array of words
            output = []
            for y in x: # for every word
                if y != "": # call decode for each  morse word if string isn't empty
                    output.append(self.decode(y))
                else:
                    output.append(" ") # add a space back in output if empty
            return "".join(output).replace("  ", " ")

    def decode(self,values):
        # a method to decode morse suing recursion 
        if values: # if not end of string
            
            # if . go to left child or is - go to right child and call recursion
            if values[0] == ".": 
                return self.leftChildNode.decode(values[1:])
            elif values[0] == "-":
                return self.rightChildNode.decode(values[1:])
            else:
                print("ERROR") # error checking
        else:
            return self.value # return the value of the object at end of string.
        

    def encode(self,word):
        # a method to handle the encoder
        word = word.upper() # cleans input
        output = [] # output string
        for letter in word: # for letters in words
            if letter == " ": # spaces for words
                output.append("  ")
                continue
            ret = self.encoder(letter) # calls the encoder function
            if ret != "": # in not empty string for error check
                output.append(ret) # append letter
                output.append(" ") # append space
        return "".join(output[:len(output)-1])
        
    def encoder(self,letter, string = ""):
        # a method to search the binary tree 
        if letter == self.value: # check is value has been found
            return string 
        else: # if not found 
            if self.leftChildNode != None: # if left child exists
                if self.leftChildNode.encoder(letter,(string + "-")) != "": # recursively check child's branch 
                    return self.leftChildNode.encoder(letter,(string + ".")) # append '.' to the string variable if found (this applies to each parent of the child's branch making the -.-.-.-. string)
            if self.rightChildNode != None: #if right child exists
                if self.rightChildNode.encoder(letter,(string + ".")) != "": # recursively check child's branch 
                    return self.rightChildNode.encoder(letter,(string + "-")) # append '-' to the string variable if found (this applies to each parent of the child's branch making the -.-.-.-. string)
            return "" # if not found in branch
        
def initTree():
    # Breadth-first array of strings 
    # alphabet = ["E","T","I","A","N","M","S","U","R","W","D","K","G","O","H","V","F","*","L","*","P","J","B","X","C","Y","Z","Q","*","*"]
    alphabet = ["E","T","I","A","N","M","S","U","R","W","D","K","G","O","H","V","F","*","L","*","P","J","B","X","C","Y","Z","Q","*","*","5",
                "4","*","3","*","¿","?","2","&","*","+","*","*","*","*","1","6","=","/","*","*","*","(","*","7","*","*","*","8","*","9","0",
                "*","*","*","*","*","*","*","*","*","*","*","*","*","_","*","*","*","*","\"","*","*",".","*","*","*","*","*","*","*","*","'",
                "*","*","-","*","*","*","*","*","*","*","*",";","!","*",")","*","*","*","¡","*",",","*","*","*","*",":","*","*","*","*","*",
                "*","*","*","*","*","*","*","*","*","*","*","$","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*",
                "*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*",
                "*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*",
                "*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*","*",
                "*","*","*","*","*","*","*"]
    
    # the stack
    queue = []

    #starting parent object  
    start = node("start")

    # add to stack
    queue.append(start)
    count = 0

    # append add the children in the binary tree.
    while count < len(alphabet)-1: # for each string in array
        parent = queue.pop(0) # grab the current parent and pop as not needed again
        childLeft = node(alphabet[count]) # make class with value from array 
        parent.setLeftChildNode(childLeft) # add to the left child of parent.
        queue.append(childLeft) # add new child to parent stack
        count += 1 # count
        childRight = node(alphabet[count]) # make class with value from array 
        parent.setRightChildNode(childRight) # add to the right child of parent.
        queue.append(childRight) # add new child to parent stack
        count += 1 # count

    return start # return parent node

def encode(value):
    # the encoder
    tree = initTree()
    return tree.encode(value)
    

def decode(value):
    # the decoder
    tree = initTree()
    return tree.words(value)